/*****************************************************************************
*  Splitting MPEG audio files at frame boundaries
*
*  Create:  27.12.2005
*  Update:  27.12.2005
*  (C)      Marcel Mller 2005
*
****************************************************************************/

#ifdef __OS2__
#define INCL_BASE
#include <os2.h>
#endif

#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#include <share.h>
#include <io.h>
#include <sys/stat.h>
#include <stdlib.h>

#undef min  // risky macros
#undef max

#include <list>
#include <memory>
#include <string>
#include <algorithm> // replace the min/max macros with side effects

using namespace std;

//#include <../../MMutil+/MMDebugMem.cpp>
string vstringf(const char* fmt, va_list va)
{  // BSD and compatible environments only:
   size_t len = vsnprintf(NULL, 0, fmt, va);
   char* cp = new char[len+1]; // political correct (do not write the internal string data structures directly)
   vsnprintf(cp, len+1, fmt, va);
   string s(cp, len);
   delete[] cp;
   return s;
}
string stringf(const char* fmt, ...)
{  va_list ap;
   va_start(ap, fmt);
   string ret = vstringf(fmt, ap);
   va_end(ap);
   return ret;
}

#define InputBufferSize 64*1024
#define OutputBufferSize 64*1024

static const double Inf = 1./0.;       // Infinity


// control vars (command line args) *****************************************
static const char*   ifname = NULL;    // input file name (required)
static const char*   ofbname = NULL;   // output file base name (optional)
static int           verbose = 1;      // verbose screen output
static int           trackno = 1;      // track number of first output file, -1 = auto-detect
static bool          autotime = false; // adjust time stamps in filename automatically
static bool          overwrite = false;// allow overwriting existing files
static list<double>  splitpoint;       // split points, +Inf identifies the end of the file, negative values are to be repeated


// Error handling ***********************************************************
enum EventClass
{  Warning = 3,
   Error
};

static const char* Class2String(EventClass cls)
{  switch (cls)
   {default:
      return "unknown";
    case Warning:
      return "Warning";
    case Error:
      return "Error";
}  }

struct Event
{  int err;
   char buf[1024];
   Event(int err, const char* fmt, ...);
};

Event::Event(int err, const char* fmt, ...)
 : err(err)
{  va_list va;
   va_start(va, fmt);
   vsprintf(buf, fmt, va);
   va_end(va);
}

// Stream helpers ***********************************************************
// buffered input stream
class byteinputstream
{private:
   #ifdef __OS2__
   HFILE          FH;
   #else
   int            FH;
   #endif
   bool           EOS;     // end of stream flag
   bool           CanSeek; // flag if the stream is seekable
   uint64_t       ByteCount;// number of bytes forwarded so far
   unsigned char* Cursor;  // head data pointer
   unsigned char* Tail;    // tail data pointer
   unsigned char* Buffer;  // data buffer

   size_t         skip(size_t len);          // skip len bytes in the stream, return remaining part on error
   void           fetch();                   // fetch as much data as possible

 public:
   byteinputstream(const char* fname);
   ~byteinputstream();
   const unsigned char* get() const          { return Cursor; }
   unsigned char  operator[](size_t i) const { return Buffer[i]; }
   uint64_t       count() const              { return ByteCount; }
   size_t         request(size_t len);       // request minimal number of bytes
   bool           forward(size_t len);       // forward the read poiter by len bytes
};

#ifdef __OS2__
// optimized version for higher throughput
byteinputstream::byteinputstream(const char* fname)
 : EOS(false)
 , ByteCount(0)
{  if (fname[0] == '-' && fname[1] == 0)
   {  FH = 0; // stdin
      CanSeek = false;
   } else
   {  ULONG ul;
      APIRET rc = DosOpen((PUCHAR)fname, &FH, &ul, 0, 0, OPEN_ACTION_FAIL_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
       OPEN_FLAGS_SEQUENTIAL|OPEN_SHARE_DENYNONE|OPEN_ACCESS_READONLY, NULL);
      if (rc != 0)
         throw Event(20, "Failed to open %s for reading. OS/2 error %lu.", fname, rc);
      ULONG type = 9999;
      DosQueryHType(FH, &type, &ul); // on error CanSeek will be false
      CanSeek = type == HANDTYPE_FILE;
   }
   Cursor = Tail = Buffer = new unsigned char[InputBufferSize];
}

byteinputstream::~byteinputstream()
{  DosClose(FH);
   delete Buffer;
}

size_t byteinputstream::skip(size_t len)
{  if (CanSeek)
   {  ULONG pos;
      APIRET rc = DosSetFilePtr(FH, len, FILE_CURRENT, &pos);
      return rc == 0 ? 0 : len;
   }
   // cannot seek -> read
   size_t rem = len;
   while (rem)
   {  ULONG n;
      APIRET rc = DosRead(FH, Buffer, InputBufferSize, &n);
      if (rc != 0)
         throw Event(27, "Failed to read input stream. OS/2 error %lu.", rc);
      if (n == 0)
      {  EOS = true;
         return rem;
      }
      if (n > rem)
      {  // more than enough
         Tail = Buffer + n;
         Cursor = Buffer + rem;
         return 0;
      }
      rem -= n;
   }
   // exact match, buffer empty
   return 0;
}

void byteinputstream::fetch()
{  unsigned char* const end = Buffer + InputBufferSize; // as much as possible
   while (Tail != end)
   {  ULONG n;
      APIRET rc = DosRead(FH, Tail, end - Tail, &n);
      if (rc != 0)
         throw Event(27, "Failed to read input stream. OS/2 error %lu.", rc);
      if (n == 0)
      {  EOS = true;
         return;
      }
      Tail += n;
   }
   return;
}

#else
// generic implementation (not that bad)
byteinputstream::byteinputstream(const char* fname)
 : EOS(false)
 , Cursor(Buffer)
 , Tail(Buffer)
 , ByteCount(0)
{  if (fname[0] == '-' && fname[1] == 0)
   {  FH = 0; // stdin
      CanSeek = false;
   } else
   {  FH = _sopen(fname, O_RDONLY|O_BINARY, SH_DENYNONE, 0);
      if (fhi == -1)
         throw Event(20, "Failed to open %s for reading. C runtime error %i.", fname, errno);
      CanSeek = true; // if a seek instruction fails, CanSeek is reset later.
   }
}

byteinputstream::~byteinputstream()
{  close(FH);
}

size_t byteinputstream::skip(size_t len)
{  if (CanSeek)
   {  if (lseek(FH, len, SEEK_CUR) != -1)
         return 0;
       else
         CanSeek = false; //
   }
   // cannot seek -> read
   size_t rem = len;
   while (rem)
   {  int n = read(FH, Buffer, InputBufferSize);
      if (n < 0)
         throw Event(27, "Failed to read input stream. C runtime error %i.", errno);
      if (n == 0)
      {  EOS = true;
         return rem;
      }
      if (n > rem)
      {  // more than enough
         Tail = Buffer + n;
         Cursor = Buffer + rem;
         return 0;
      }
      rem -= n;
   }
   return 0;
}

void byteinputstrem::fetch()
{  unsigned char* const end = Buffer + InputBufferSize; // as much as possible
   while (Tail != end)
   {  int n = read(FH, Tail, end - Tail);
      if (n < 0)
         throw Event(27, "Failed to read input stream. C runtime error %i.", errno);
      if (n == 0)
      {  EOS = true;
         return;
      }
      Tail += n;
   }
   return;
}

#endif

size_t byteinputstream::request(size_t len)
{  if (len > InputBufferSize)
      throw Event(28, "Internal error: can't request mor data (%u) than the buffer size (%u).", len, InputBufferSize);
   if (!EOS)
   {  if (Tail == Cursor)
         Cursor = Tail = Buffer;
       else if (Cursor - Buffer > InputBufferSize - len)
      {  // move data, can't help
         memmove(Buffer, Cursor, Tail - Cursor);
         Tail -= Cursor - Buffer;
         Cursor = Buffer;
      }
      fetch();
   }
   return Tail - Cursor;
}

bool byteinputstream::forward(size_t len)
{  if (Cursor + len <= Tail)
   {  Cursor += len;
      ByteCount += len;
      return true;
   }
   Cursor = Tail = Buffer; // invalidate Buffer
   if (EOS)
      return false;
   size_t rem = skip(len - (Tail - Cursor));
   ByteCount += len - rem;
   return rem == 0;
}

// buffered output stream
class byteoutputstream
{  FILE*          FH;
   uint64_t       ByteCount;// number of bytes forwarded so far
 public:
   byteoutputstream() : FH(NULL) {}
   ~byteoutputstream() { reset(); }
   void           setfile(const char* fname, size_t filesize = 0);
   bool           active() const             { return FH != NULL; }
   uint64_t       count() const              { return ByteCount; }
   void           reset();
   void           write(const void* data, size_t len);
};

void byteoutputstream::setfile(const char* fname, size_t length)
{  if (FH)
      fclose(FH);
   #if defined(__OS2__) && defined(__GCC__)
   int fh = open(fname, overwrite ? O_WRONLY|O_BINARY|O_CREAT|O_TRUNC|O_SIZE : O_WRONLY|O_BINARY|O_CREAT|O_EXCL|O_SIZE, S_IREAD|S_IWRITE, length);
   #else
   int fh = open(fname, overwrite ? O_WRONLY|O_BINARY|O_CREAT|O_TRUNC : O_WRONLY|O_BINARY|O_CREAT|O_EXCL, S_IREAD|S_IWRITE);
   #endif
   if (fh == -1)
      throw Event(21, "Failed to open %s for writing. C runtime error %i.", fname, errno);
   FH = fdopen(fh, "wb");
   if (FH == NULL)
      throw Event(28, "Failed to open internal file handle (fdopen). C runtime error %i.", fname, errno);
   setvbuf(FH, NULL, _IOFBF, OutputBufferSize);
   ByteCount = 0;
}

void byteoutputstream::reset()
{  if (FH)
   {  fclose(FH);
      FH = NULL;
   }
   ByteCount = 0;
}

void byteoutputstream::write(const void* data, size_t len)
{  if (FH && fwrite(data, 1, len, FH) != len)
      throw Event(27, "Failed to write output file. C runtime error %i.", errno);
   ByteCount += len;
}

// file name generator ******************************************************

class namefactory
{protected:
   string   Name;
   unsigned StartIndex;

 public:
   namefactory(const string& fname);
   virtual  ~namefactory() {}
   virtual string makename(int seconds) = 0;
};

namefactory::namefactory(const string& fname)
 : Name(fname)
{  unsigned p = Name.find_last_of("\\/:");
   StartIndex = p == string::npos ? 0 : p+1;
}

class tracknamefactory : public namefactory
{protected:
   int      Track;
   unsigned InsertionPoint;
 public:
   tracknamefactory(const string& fname, int starttrack);
   string   makename(int seconds);
};

tracknamefactory::tracknamefactory(const string& fname, int starttrack)
 : namefactory(fname)
 , Track(starttrack)
{  // search insertion point
   InsertionPoint = Name.rfind('.');
   if (InsertionPoint == string::npos || InsertionPoint < StartIndex)
      InsertionPoint = Name.length();
}

string tracknamefactory::makename(int)
{  return Track == 0 ? Name : string(Name).insert(InsertionPoint, stringf("%02i", Track++));
}


class autotimenamefactory : public namefactory
{protected:
   time_t   StartTime;
   unsigned TimePos;
   unsigned TimeDigits;
   unsigned DatePos;
   unsigned DateDigits;
 private:
   unsigned FindDigits(unsigned len, unsigned start);
   unsigned FindDigits(unsigned len)                  { return FindDigits(len, StartIndex); }
   int      ExtractInt(unsigned pos, unsigned len);
   void     OverlayInt(unsigned pos, unsigned len, int value);

 public:
   autotimenamefactory(const string& fname);
   string   makename(int seconds);
};

unsigned autotimenamefactory::FindDigits(unsigned len, unsigned start)
{  for (;;)
   {  unsigned p = Name.find_first_of("0123456789", start);
      if (p == string::npos)
         return string::npos;
      start = Name.find_first_not_of("0123456789", p+1);
      if (start == string::npos)
         return Name.length() - p >= len ? p : string::npos;
      if (start - p >= len)
         return p;
   }
}

int autotimenamefactory::ExtractInt(unsigned pos, unsigned len)
{  int r = 0;
   while (len--)
      r = 10*r + (Name[pos++] - '0');
   return r;
}

void autotimenamefactory::OverlayInt(unsigned pos, unsigned len, int value)
{  pos += len;
   while (len--)
   {  Name[--pos] = '0' + value%10;
      value /= 10;
   }
}

static time_t mkgmt(struct tm* t)
{  static const int monsum[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};
   static const int monlen[] = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
   unsigned int year70 = t->tm_year - 70;
   bool leapyear = (t->tm_year & 3) == 0;
   //printf("%i %i %i %i %i %i  %i %i\n", t->tm_year, t->tm_mon, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, year70, leapyear);
   // check
   if ( (t->tm_sec|t->tm_min|t->tm_hour|t->tm_mon) < 0
      || t->tm_sec > 59 || t->tm_min > 59 || t->tm_hour > 23
      || year70 >= 136U || t->tm_mon > 11 || t->tm_mday < 1 || t->tm_mday > monlen[t->tm_mon]
      || (!leapyear && t->tm_mday == 29 && t->tm_mon == 1) )
      return (time_t)-1;
   // fill fields
   t->tm_yday = t->tm_mday-1 + monsum[t->tm_mon] + (leapyear && t->tm_mon >= 2);
   time_t days = t->tm_yday + 365*year70 + (year70+1)/4;
   t->tm_wday = (days+4) % 7;
   t->tm_isdst = false;
   return t->tm_sec + 60*t->tm_min + 3600*t->tm_hour + 86400U * days;
}

autotimenamefactory::autotimenamefactory(const string& fname)
 : namefactory(fname)
{  // analyze filename
   DateDigits = 8;
   TimeDigits = 6;
   DatePos = FindDigits(8+6);
   if (DatePos != string::npos)
      TimePos = DatePos + 8;
    else
   {  DatePos = FindDigits(8+4);
      if (DatePos != string::npos)
      {  TimePos = DatePos + 8;
         TimeDigits = 4;
      } else
      {  DatePos = FindDigits(8);
         if (DatePos == string::npos)
         {  DatePos = FindDigits(6);
            DateDigits = 6;
         }
         TimePos = FindDigits(6, DatePos + DateDigits +1);
         if (TimePos == string::npos)
         {  TimePos = FindDigits(4, DatePos + DateDigits +1);
            TimeDigits = 4;
            if (TimePos == string::npos)
               throw Event(30, "Failed to extract time stamp from %s.", Name.c_str());
         }
      }
   }
   if (verbose > 2)
      printf("Found timestamp at:\n"
         "Date: Character %i - %i: %.*s\n"
         "Time: Character %i - %i: %.*s\n"
         , DatePos, DatePos+DateDigits-1, DateDigits, Name.data()+DatePos, TimePos, TimePos+TimeDigits-1, TimeDigits, Name.data()+TimePos);
   // extract time
   struct tm start;
	start.tm_sec = ExtractInt(TimePos +4, TimeDigits -4);
	start.tm_min = ExtractInt(TimePos +2, 2);
	start.tm_hour = ExtractInt(TimePos, 2);
	start.tm_mday = ExtractInt(DatePos + DateDigits -2, 2);
	start.tm_mon = ExtractInt(DatePos + DateDigits -4, 2) -1;
	start.tm_year = ExtractInt(DatePos, DateDigits-4);
   if (DateDigits == 8)
      start.tm_year -= 1900;
    else if (start.tm_year < 70)
      start.tm_year += 100;
	start.tm_isdst = false;
	// store as unix time stamp
   StartTime = mkgmt(&start);
   if (StartTime == (time_t)-1)
      throw Event(30, "The extracted time stamp %.*s %.*s is not valid.", DateDigits, Name.data()+DatePos, TimeDigits, Name.data()+TimePos);
}

string autotimenamefactory::makename(int seconds)
{  time_t t = StartTime + seconds;
   struct tm* tp = gmtime(&t);
   OverlayInt(TimePos +4, TimeDigits-4, tp->tm_sec);
   OverlayInt(TimePos +2, 2, tp->tm_min);
   OverlayInt(TimePos, 2, tp->tm_hour);
   OverlayInt(DatePos + DateDigits -2, 2, tp->tm_mday);
   OverlayInt(DatePos + DateDigits -4, 2, tp->tm_mon +1);
   OverlayInt(DatePos, DateDigits-4, tp->tm_year +1900);
   return Name;
}

// mpeg tables **************************************************************
static const char sVersion[4][4] = {"2.5", {'','0','1',''}, "2", "1"};
static const char sLayer[4][4] = {{'','0','0',''}, "III", "II", "I"};
static const char sYesNo[2][4] = {"no", "yes"};
static const int iBitRate[2][4][16] = {
{  {-2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2},
   {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1},
   {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, -1},
   {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1}
},
{  {-2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2},
   {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1},
   {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1},
   {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1}
} };
static const int iSampleRate[4][4] =
{  {11025, 12000, 8000, -1},
   {-2, -2, -2, -2},
   {22050, 24000, 16000, -1},
   {44100, 48000, 32000, -1}
};
static const char sMode[4][2] = {"S", {'J','S'}, {'D','C'}, {'S','C'}};
static const char sExtMode[2][4][8] =
{  {"4-31", "8-31", "12-31", "16-31"},
   {"off off", "on  off", "off on", "on  on"}
};
static const char sSExtMode[2][4][8] =
{  {"IS 4-31", "IS 8-31", {'I','S',' ','1','2','-','3','1'}, {'I','S',' ','1','6','-','3','1'}},
   {"normal", "IS", "MS-S", "IS+MS-S"}
};
static const char sCopyright[4][5] = {"none", "none", {'o','r','i','g','.'}, "copy"};
static const char sEmphasis[4][5] = {"none", {'5','0','/','1','5'}, "10", "CCIT"};

static const unsigned iSideInfoLen[4] = {9, 17, 17, 32};


// mpeg structs (little endian) *********************************************
struct MPEG_Frame_Header
{  // layout
   signed   FrameSync1    : 8;
   unsigned CRC           : 1;
   unsigned Layer         : 2;
   unsigned Version       : 2;
   signed   FrameSync2    : 3;
   unsigned Private       : 1;
   unsigned Padding       : 1;
   unsigned SamplingRate  : 2;
   unsigned BitRate       : 4;
   unsigned Emphasis      : 2;
   unsigned Copyright     : 2;
   unsigned ModeExtension : 2;
   unsigned Mode          : 2;
   // methods
   int      GetLayer() const           { return 4 - Layer; }
   int      GetSamplingRate() const    { return iSampleRate[Version][SamplingRate]; }
   const int* GetBitRates() const      { return iBitRate[Version != 3][Layer]; }
   int      GetBitRate(const int* brv) const { return brv[BitRate]; }
   int      GetBitRate() const         { return iBitRate[Version != 3][Layer][BitRate]; }
   size_t   GetFrameLen(int br) const;
   size_t   GetFrameLen() const        { return GetFrameLen(GetBitRate()); }
   bool     HasCRC() const             { return !CRC; }
   bool     Has2Channels() const       { return Mode != 3 && Layer == 1; }
   size_t   GetFrameSamples() const    { return Layer == 3 ? 384U : 1152U; }
   size_t   GetGranules() const        { return (Version == 3) +1; }
};

size_t MPEG_Frame_Header::GetFrameLen(int br) const
{  int sr = iSampleRate[3][SamplingRate];
   return (br <= 0 || sr <= 0 || Version == 1)
    ? 0 : (Layer == 3)
      ? (12000 * br / sr + Padding) << 2
      : ((144000 * br) << (Version != 3)) / sr + Padding; // I _really_ don't know why the >> operation is needed.
}


// XING VBR header funcs ****************************************************
/*static int ExtractI4(const unsigned char* &buf)
{  // big endian extract
   int x = (((buf[0] << 8) + buf[1] << 8) + buf[2] << 8) + buf[3];
   buf += 4;
   return x;
}

static void StoreI4(unsigned char* &buf, int i)
{  // big endian store
   buf[3] = (unsigned char)i;
   i >>= 8;
   buf[2] = (unsigned char)i;
   i >>= 8;
   buf[1] = (unsigned char)i;
   i >>= 8;
   buf[0] = (unsigned char)i;
   buf += 4;
}

#define FRAMES_FLAG     0x0001
#define BYTES_FLAG      0x0002
#define TOC_FLAG        0x0004
#define VBR_SCALE_FLAG  0x0008

#define FRAMES_AND_BYTES (FRAMES_FLAG | BYTES_FLAG)
// structure to receive extracted header
// toc may be NULL
struct XHEADDATA
{  int h_id;       // from MPEG header, 3=MPEG1, 2=MPEG2, 0=MPEG2.5
   int h_mode;     // from MPEG header, channel mode, 3=mono
   int samprate;   // determined from MPEG header
   int flags;      // from Xing header data
   int frames;     // total bit stream frames from Xing header data
   int bytes;      // total bit stream bytes from Xing header data
   int vbr_scale;  // encoded vbr scale from Xing header data
   unsigned char toc[100];// toc_buffer[100]

   static auto_ptr<XHEADDATA> Get(const unsigned char *buf);
   void Put(unsigned char *buf) const;
   void Recalculate(size_t _frames, size_t _bytes, size_t* frameloc, size_t offset = 0);
   size_t Offset() const      { return 4 + iSideInfoLen[((h_id == 3) << 1) + (h_mode != 3)]; }
   size_t Length() const;
   #ifdef DEBUG
   void Print() const;
   #endif
};

#ifdef DEBUG
void XHEADDATA::Print() const
{  printf("*** XING-Header ***\n"
          "ID            %i\n"
          "Samplingrate  %i\n"
          "Flags         %X\n"
          "Frames        %i\n"
          "Bytes         %i\n"
          "VBR scale     %i\n",
    h_id, samprate, flags, frames, bytes, vbr_scale);
   for (int i = 0; i < 100; ++i)
      printf("%3i ", toc[i]);
   puts("");
}
#endif

auto_ptr<XHEADDATA> XHEADDATA::Get(const unsigned char *buf)
// return 0=fail, 1=success
// X   structure to receive header data (output)
// buf bitstream input
{  auto_ptr<XHEADDATA> X(new XHEADDATA);

   // get Xing header data
   X->flags = 0;     // clear to null incase fail

   // get selected MPEG header data
   X->h_id     = (buf[1] >> 3) & 3;
   X->samprate = iSampleRate[X->h_id][(buf[2] >> 2) & 3];
   X->h_mode   = (buf[3] >> 6) & 3;

   // determine offset of header
   buf += X->Offset();

   if (buf[0] != 'X' || buf[1] != 'i' || buf[2] != 'n' || buf[3] != 'g')
      return auto_ptr<XHEADDATA>();              // fail, header not found
   buf+=4;

   X->flags = ExtractI4(buf);// get flags

   if (X->flags & FRAMES_FLAG)
      X->frames   = ExtractI4(buf);
   if (X->flags & BYTES_FLAG)
      X->bytes = ExtractI4(buf);

   if (X->flags & TOC_FLAG)
      memcpy(X->toc, buf, 100), buf+=100;

   X->vbr_scale = -1;
   if (X->flags & VBR_SCALE_FLAG)
      X->vbr_scale = ExtractI4(buf);

   #ifdef DEBUG
   X->Print();
   #endif

   return X;       // success
}

void XHEADDATA::Put(unsigned char *buf) const
// X   structure to header data (input)
// buf bitstream input/output
{  // determine offset of header
   buf += Offset();

   buf[0] = 'X';
   buf[1] = 'i';
   buf[2] = 'n';
   buf[3] = 'g';
   buf+=4;

   StoreI4(buf, flags);

   if (flags & FRAMES_FLAG)
      StoreI4(buf, frames);
   if (flags & BYTES_FLAG)
      StoreI4(buf, bytes);

   if (flags & TOC_FLAG)
      memcpy(buf, toc, 100), buf+=100;

   if (flags & VBR_SCALE_FLAG)
      StoreI4(buf, vbr_scale);
}

void XHEADDATA::Recalculate(size_t _frames, size_t _bytes, size_t* frameloc, size_t offset)
{  frames = _frames;
   bytes = _bytes;
   flags = FRAMES_FLAG | BYTES_FLAG | TOC_FLAG;
   // recreate TOC
   offset -= frameloc[0]; // gives an overflow, this is intended
   for (int i = 0; i < 100; ++i)
      toc[i] = (long long)(frameloc[(i * frames + 50) /100] + offset) * 256 / bytes;
   #ifdef DEBUG
   Print();
   #endif
}

size_t XHEADDATA::Length() const
{  size_t len = 4;

   if (flags & FRAMES_FLAG)
      len += 4;
   if (flags & BYTES_FLAG)
      len += 4;
   if (flags & TOC_FLAG)
      len += 100;
   if (flags & VBR_SCALE_FLAG)
      len += 4;

   return len;
}
*/


static void frameerror(EventClass cls, unsigned long frame, uint64_t pos, const char* fmt, ...)
{  fprintf(stderr, "%s at frame %lu (@ %Lx): ", Class2String(cls), frame+1, pos);
   va_list va;
   va_start(va, fmt);
   vfprintf(stderr, fmt, va);
   va_end(va);
   fputc('\n', stderr);
}


static void printframeinfo(long frames, double bitrate, const MPEG_Frame_Header hdr)
{  puts("frames  ver. layer  CRC  kbps   sampl. pad. mode  IS  MS-S  (C)   emph.");
   printf("%-8lu%-5.4s%-7.4s%-5.4s%5.1f  %-5i  %-5.4s%-6.2s%-10.8s%-6.5s%.5s\n",
    frames, sVersion[hdr.Version], sLayer[hdr.Layer], sYesNo[!hdr.CRC],
    bitrate, max(iSampleRate[hdr.Version][hdr.SamplingRate], 0),
    sYesNo[hdr.Padding], sMode[hdr.Mode],
    hdr.Mode == 1 ? sExtMode[hdr.Layer <= 1][hdr.ModeExtension] : "",
    sCopyright[hdr.Copyright], sEmphasis[hdr.Emphasis]);
}

static string formattime(double s)
{  long l = (long)s;
   s -= l;
   if (s != 0)
   {  // frational seconds
      if (l >= 3600)
         return stringf("%d:%02d:%05.2f", l/3600, l/60%60, s);
      if (l >= 60)
         return stringf("%d:%05.2f", l/60, s);
      return stringf("%.2f", s);
   } else
   {  if (l >= 3600)
         return stringf("%d:%02d:%02d", l/3600, l/60%60, l%60);
      if (l >= 60)
         return stringf("%d:%02d", l/60, l%60);
      return stringf("%d", l);
   }
}


// command line parsing funcs ***********************************************
static int intoption(const char* arg, const char* context)
{  size_t n = 0;
   int r;
   sscanf(arg, "%i%n", &r, &n);
   if (n != strlen(arg))
      throw Event(42, "Integer constant expected at %s.", context);
   return r;
}

static double timeoption(const char* arg, const char* context)
{  double t = 0;
   double d;
   size_t n;
   for (;;)
   {  const char* pc = strchr(arg, ':');
      if (pc == NULL)
         break;
      if (sscanf(arg, "%lf%n", &d, &n) != 1 || n != (size_t)(pc-arg))
         throw Event(42, "Time constant %.*s must be a number at %s.", pc-arg, arg, context);
      t = 60*t + d;
      arg = pc+1;
   }
   if (sscanf(arg, "%lf%n", &d, &n) != 1 || n != strlen(arg))
      throw Event(42, "Time constant %s must be a number at %s.", arg, context);
   return 60*t + d;
}

static void parseoption(const char* cp)
{  switch (tolower(cp[0]))
   {default:
      throw Event(44, "Illegal option %s.", cp);

    case 'v':
      verbose = 2;
      return;
    case '_':
      verbose = 3;
      return;
    case 'q':
      verbose = 0;
      return;

    case 'n':
      if (cp[1] == 0)
         trackno = -1; // auto-detect
       else
         trackno = intoption(cp+1, cp);
      return;
    case 'b':
      ofbname = cp+1;
      return;
    case 't':
      autotime = true;
      return;
    case 'o':
      overwrite = true;
      return;
   }
}


int _cdecl main(int argc, char** argv)
{  // ***** parse command line ...
   if (argc < 2)
   {  fprintf(stderr, "usage: %s [options] sourcefile start [split [split]] end\n"
         "Positions are given in seconds or minutes:seconds or hours:minutes:seconds.\n"
         "If they are preceeded with + it counts relative to the previous one.\n"
         "If the last point is preceeded by ++ the source is splitted every xxx\n"
         "hours/minuts/seconds from the previous split point until the end of the file.\n"
         "The Last Position may be * whitch means up to the end of the source file.\n"
         "The destination files are named automatically file01, file02, ... unless the\n"
         "/t option is used.\n"
         "options: /n#              start with destination file number # instead of 01.\n"
         "         /n               auto detect Track number.\n"
         "         /b$              change basename or path for destination files to $.\n"
         "         /t               Read and adjust filenames with [yy]yymmdd and hhmm[ss].\n"
         "         /o               Overwrite existing output files.\n"
         , argv[0]);
      return 10;
   }
   try
   {  for (int i = 1; i < argc; ++i)
      {  char* cp = argv[i];
         if (cp[0] == '/') // option
            parseoption(cp+1);
          else if (!ifname) // first non-option => input filename
            ifname = cp;
          else if (cp[0] == '*' && cp[1] == 0) // splitpoint: to the end
            splitpoint.push_back(+Inf);
          else if (cp[0] == '+' && cp[1] == '+') // autosplit
         {  double val = timeoption(cp+2, cp);
            if (val == 0)
               throw Event(42, "Autoincrement cannot be zero at %s.", cp);
            if (splitpoint.size() == 0)
               splitpoint.push_back(0); // automatically prepend start of file
            splitpoint.push_back(-val);
         } else if (cp[0] == '+') // relative splitpoint
         {  double val = timeoption(cp+1, cp);
            if (splitpoint.size())
            {  if (splitpoint.back() < 0)
                  throw Event(42, "Cannot use relative time span %s after iterative split point ++%s.", cp, formattime(-splitpoint.back()).c_str());
               val += splitpoint.back();
            }
            splitpoint.push_back(val);
         } else // absolute splitpoint
            splitpoint.push_back(timeoption(cp, cp));
      }

      // ***** check parameter constraints ...
      if (!ifname)
         throw Event(40, "Input filename required.");
      if (splitpoint.size() < 2)
         throw Event(41, "At least a starting and a end point is required.");
      double last2 = 0;
      double last = 0;
      for (list<double>::iterator i = splitpoint.begin(); i != splitpoint.end(); ++i)
      {  if (last == +Inf)
            throw Event(34, "The * (end of file) split point must be the last one.");
         if (last < 0)
         {  if (*i < 0)
               throw Event(34, "Two iterative split points in sequence (++%s and ++%s).", formattime(-last).c_str(), formattime(*i).c_str());
            swap(last, last2);
         }
         if (*i >= 0 && *i != +Inf && *i < last)
            throw Event(34, "Negative time span from %s to %s.", formattime(last).c_str(), formattime(*i).c_str());
         if (last2 < 0 && -last2 >= *i - last)
         {  fprintf(stderr, "Warning: ignoring iterative split point ++%s because the next absolute split point %s is closer than one iteration to %s.\n"
             , formattime(-last2).c_str(), formattime(*i).c_str(), formattime(last).c_str());
            splitpoint.erase(--list<double>::iterator(i));
         }
         last2 = last;
         last = *i;
      }
      /*if (autotime && splitpoint.front() == 0. && ofbname == NULL)
         throw Event(34, "The first output file is the same as the input file.");*/

      if (verbose > 2) // echo command line
      {  for (int i = 1; i < argc; ++i)
            printf("%s ", argv[i]);
         putchar('\n');
      }

      // initialization
      byteinputstream bi(ifname);   // open input file
      byteoutputstream bo;          // initialized later
      string ofname;
      if (ofbname == NULL)          // destination file template
      {  ofname = ifname;
         unsigned p = ofname.find_last_of("\\/:");
         if (p != string::npos)
            ofname.erase(0, p+1);
      } else
      {  ofname = ofbname;
         unsigned p = ofname.find_last_of("\\/:");
         if (p == ofname.length()-1) // found at the last character
         {  // append filename from ifname
            p = string(ifname).find_last_of("\\/:");
            ofname += p == string::npos ? ifname : ifname + p +1;
         }
      }
      auto_ptr<namefactory> nf(autotime ? (namefactory*)new autotimenamefactory(ofname) : new tracknamefactory(ofname, trackno));

      MPEG_Frame_Header hdr_OK;
      list<double>::iterator nextsplit = splitpoint.begin();
      unsigned long nextsplitframe = (unsigned long)-1;     // initialized later

      if (verbose > 0)
         printf("Reading input %s ...\n", ifname);

      // skip V2.x tag if any
      if (bi.request(10) >= 10 && memcmp(bi.get(), "ID3", 3) == 0 && bi[3] != 0xff && bi[4] != 0xff && (bi[6]|bi[7]|bi[8]|bi[9]) < 0x80)
      {  size_t ID3v2Tag = 10 + bi[9] + ((bi[8] + (bi[7] + (bi[6] << 7) << 7)) << 7);
         if (!bi.forward(ID3v2Tag))
            throw Event(27, "Skip of V2.x tag failed.");
         fprintf(stderr, "Information: discarding ID3V2 tag of %u bytes at start of file.\n", ID3v2Tag);
      }

      if (verbose > 0)
      {  puts("from\tto\tfile");
         if (splitpoint.front() > 0.)
            printf("0\t%s\t*** skip ***\n", formattime(splitpoint.front()).c_str());
      }

      unsigned long frame = 0;
      unsigned long outframe = 0;
      bool terminate = false;
      do
      {  // read MPEG Header
         if (bi.request(4) < 4)
            break;
         // check header
         #define hdr (*(MPEG_Frame_Header*)bi.get())
         if ((hdr.FrameSync1 & hdr.FrameSync2) != -1)
         {  if (bi.request(128) >= 128 && memcmp(bi.get(), "TAG", 3) == 0) // V1.x Tag
            {  bi.forward(128);
               fprintf(stderr, "Information: discarding ID3V1.x tag at %Lx.\n", bi.count());
               continue;
            }
            frameerror(Error, frame, bi.count(), "Illegal frame sync 0x%02x%x", hdr.FrameSync1&0xFF, hdr.FrameSync2&7);
          resync:
            bi.forward(1);
          resync2:
            // try resync
            size_t n = bi.request(4);
            if (n < 4)
               break;
            const unsigned char* p = (const unsigned char*)memchr(bi.get(), 0xFF, n);
            if (p == NULL)
            {  bi.forward(n);
               goto resync2;
            } else
               bi.forward(p - bi.get());
            // potential match is before us
            if (bi[1] < 0xE0 || bi[2] >= 0xC0 || hdr.SamplingRate == 3)
               goto resync;
            // found possible sync
            fprintf(stderr, "Resync at %Lx\n", bi.count());
            continue;
         }

         int bitrate = hdr.GetBitRate();
         switch (bitrate)
         {case -1:
            frameerror(Error, frame, bi.count(), "Bitrate bad (%x)", bitrate);
          case -2:
            bitrate = 0;
          default:;
         }
         int samprate = hdr.GetSamplingRate();
         switch (samprate)
         {case -1:
            frameerror(Error, frame, bi.count(), "Sampling rate bad (%x)", samprate);
          case -2:
            samprate = 0;
          default:;
         }
         size_t framelen = hdr.GetFrameLen(bitrate);

         if (framelen > 4)
         {  // frame seems to be valid
            hdr_OK = hdr;
            #ifdef DEBUG
            if (verbose > 2)
               printf("Frame: %Lx, %08x\n", bi.count(), *(int*)&hdr_OK);
            #endif
            // request frame data
            size_t len = bi.request(framelen);
            if (len < framelen)
            {  frameerror(Warning, frame, bi.count(), "Incomplete frame. Expected frame length %u but only %u bytes in stream left.", framelen, len);
               framelen = len; // truncate
            }

            // check split for points
            for(;;)
            {  if (nextsplitframe == (unsigned long)-1)
               {  // get last
                  list<double>::iterator i = nextsplit;
                  --i;
                  double last = i == splitpoint.end() ? 0 : *i;
                  if (*nextsplit < 0) // autoincrement
                  {  // get next
                     i = nextsplit;
                     ++i;
                     if (i != splitpoint.end() && *i <= last - *nextsplit)
                        // end of autoincrement section => remove the autoincrement marker
                        nextsplit = splitpoint.erase(nextsplit);
                      else
                        // insert previos splitpoint - current value befor the current (autoincrement) value.
                        nextsplit = splitpoint.insert(nextsplit, last - *nextsplit);
                  }
                  // now *nextsplit is a positive number or +Inf
                  nextsplitframe = *nextsplit == +Inf ? LONG_MAX : (long)(*nextsplit/hdr.GetFrameSamples()*samprate + .5); // calc split point in frames
                  if (verbose > 0 && bo.active())
                  {  if (*nextsplit == +Inf)
                        printf("%s\tto end\t%ld\t%s\n", formattime(last).c_str(), nextsplitframe, ofname.c_str());
                      else
                        printf("%s\t%s\t%ld\t%s\n", formattime(last).c_str(), formattime(*nextsplit).c_str(), nextsplitframe, ofname.c_str());
                  }
               }
               if (frame < nextsplitframe)
                  break;
               // next file
               if (verbose > 1 && nextsplit != splitpoint.begin())
                  printframeinfo(outframe, bo.count() * .8 / hdr_OK.GetFrameSamples() / outframe / 4 * (hdr_OK.GetSamplingRate() / 25), hdr_OK);
               double lastsplit = *nextsplit;
               if (++nextsplit == splitpoint.end()) // finito?
               {  bo.reset();
                  terminate = true;
                  nextsplitframe = LONG_MAX; // no more
               } else
               {  ofname = nf->makename((int)(lastsplit + .5));
                  bo.setfile(ofname.c_str());
                  nextsplitframe = (unsigned long)-1; // invalidate
               }
               outframe = 0;
            }
            // copy data if any
            bo.write(bi.get(), framelen);

            // next frame
            ++frame;
            ++outframe;
            bi.forward(framelen);
         }
      } while (!terminate);

      if (!frame)
         throw Event(10, "Warning: File %s contains no valid MPEG Frames.", ifname);

      if (nextsplitframe != LONG_MAX && nextsplitframe > frame && splitpoint.back() >= 0) // not in autosplit mode
         throw Event(2, "Warning: Split point at %s (frame #%lu) is beyond the end of the file.", formattime(*nextsplit).c_str(), nextsplitframe);
      if (verbose > 1 && !terminate)
         printframeinfo(outframe, bo.count() * .8 / hdr_OK.GetFrameSamples() / outframe / 4 * (hdr_OK.GetSamplingRate() / 25), hdr_OK);
      if (verbose > 0 && terminate && bi.request(4) >= 4)
         printf("%s\tto end\t*** ignore ***\n", formattime(splitpoint.back()).c_str());

      return 0;

   } catch (const Event& err)
   {  fprintf(stderr, "%s\n", err.buf);
      return err.err;
   }
}
